#include <tw2dcore.h>
#include <twcomp.h>
#include <stdio.h>



/////////////////////////////////////////////////////////////////////
//------------------------ TWCOLOR --------------------------------//
/////////////////////////////////////////////////////////////////////
BYTE TWCOLOR::GetComponent(const DWORD & col,DWORD mask)
{
	if((!col) || (!mask)) return 0;
	DWORD tmp;
	tmp=col&mask;
	while(!(mask&0x1)) 
	{
		tmp>>=1;
		mask>>=1;
	}
	tmp*=255;
	tmp/=mask;
	return (BYTE)(tmp);
}

void TWCOLOR::SetComponent(DWORD & col,DWORD mask,BYTE comp)
{
	if((!mask) || (!comp)) return;
	DWORD m2=mask;
	DWORD tmp=comp;
	int i=0;
	
	while(!(mask&0x1)) 
	{
		++i;
		mask>>=1;
	}
	tmp*=mask;
	tmp/=255;
	tmp<<=i;
	
	col&=~m2;
	col|=tmp;
}

void TWCOLOR::SetRGBA32(const DWORD & c)
{
	b=GetComponent(c,bmask);
	g=GetComponent(c,gmask);
	r=GetComponent(c,rmask);
	a=GetComponent(c,amask);
}

DWORD TWCOLOR::GetRGBA32() const
{
	DWORD ret=0;
	SetComponent(ret,rmask,r);
	SetComponent(ret,gmask,g);
	SetComponent(ret,bmask,b);
	SetComponent(ret,amask,a);
	return ret;
}


void TWCOLOR::BGRToRGB()
{
	BYTE t=b;
	b=r;
	r=t;
}

WORD TWCOLOR::GetRGB16() const
{
	register DWORD res;	
	SetComponent(res,rmask,r);
	SetComponent(res,gmask,g);
	SetComponent(res,bmask,b);
	return (WORD)res;
}


void TWCOLOR::SetRGB16(WORD c)
{
	r=GetComponent(c,rmask);
	g=GetComponent(c,gmask);
	b=GetComponent(c,bmask);	
}

void TWCOLOR::SetRGB24(const BYTE * c)
{
	DWORD tmp=c[0]|(c[1]<<8)|(c[2]<<16);
	r=GetComponent(tmp,rmask);
	g=GetComponent(tmp,gmask);
	b=GetComponent(tmp,bmask);	
}

DWORD TWCOLOR::GetYUV32() const
{
	return GetYUV24()|(((DWORD)a)<<24);
}

DWORD TWCOLOR::GetYUV24() const
{
	register DWORD res=0;
	res=(DWORD)GetV()|((DWORD)GetU()<<8)|((DWORD)GetY()<<16)|((DWORD)a<<24);
	return res;
}


BYTE TWCOLOR::GetY() const
{
	double y=0.299*(double)r+0.587*(double)g+0.114*(double)b;
	//if(y>255) y=255;
	return (BYTE)y;
}

BYTE TWCOLOR::GetU() const
{
	double u=-0.1687*(double)r-0.3313*(double)g+0.5*(double)b+128;
	return (BYTE)u;
}

BYTE TWCOLOR::GetV() const
{
	double v=0.5*(double)r-0.4187*(double)g-0.0813*(double)b+128;
	return (BYTE)v;
}

void TWCOLOR::SetYUV(BYTE y,BYTE u,BYTE v,BYTE a)
{
	DWORD t=v|(DWORD)u<<8|(DWORD)y<<16|(DWORD)a<<24;
	SetYUV32(t);
}

void TWCOLOR::SetYUV32(DWORD c)
{
	double y,u,v,t;
	v=(double)(c&0xFF);
	u=(double)((c&0xFF00)>>8);
	y=(double)((c&0xFF0000)>>16);
	a=(BYTE)((c&0xFF000000)>>24);

	u-=128;
	v-=128;
	t=y - 0.0009267*u + 1.4016868*v;//y+1.402*v;
	if(t<0) t=0;
	if(t>255) t=255;
	r=(BYTE)t;
	t=y - 0.3436954*u - 0.7141690*v;//y-0.34414*u-0.71414*v;
	if(t<0) t=0;
	if(t>255) t=255;
	g=(BYTE)t;
	t=y + 1.7721604*u + 0.0009902*v;//y+1.772*u;
	if(t<0) t=0;
	if(t>255) t=255;
	b=(BYTE)t;
}

void TWCOLOR::SetYUV24(DWORD c)
{
	SetYUV32(c);
	a=0;
}

/////////////////////////////////////////////////////////////////////
//----------------------- TWPICTURE -------------------------------//
/////////////////////////////////////////////////////////////////////
TWPICTURE::TWPICTURE() : TWOBJECT("TWPICTURE"), TWLOADABLEOBJECT("Image Loader")
{
	buffer=workbuf=NULL;
	palette=NULL;
	Release();
}

TWPICTURE::~TWPICTURE()
{
	Release();
}

void *& TWPICTURE::CreateBuffer()
{
	TWOBJECT::FreeBuffer(&buffer);
	bufsize=width*height*(pixelformat&0xFF);
	if(!bufsize) return buffer;
	if(bufsize%8) bufsize+=8;
	bufsize>>=3;
	switch(pixelformat&0xFF)
	{
	case 16:
		rmask=0x1F;
		gmask=0x7E0;
		bmask=0xF800;
		amask=0;
		break;
	case 24:
		rmask=0xFF;
		gmask=0xFF00;
		bmask=0xFF0000;
		amask=0;
		break;
	case 32:
		rmask=0xFF;
		gmask=0xFF00;
		bmask=0xFF0000;
		amask=0xFF000000;
		break;
	}
	TWOBJECT::CreateBuffer(&buffer,bufsize);
	return buffer;
}


void * TWPICTURE::CreateWorkBuffer()
{
	if(!buffer) return NULL;
	TWOBJECT::CreateBuffer((void **)&workbuf,width*height*sizeof(DWORD));
	return workbuf;
}


void TWPICTURE::Release()
{
	TWOBJECT::FreeBuffer((void**)&buffer);
	TWOBJECT::FreeBuffer((void**)&workbuf);
	TWOBJECT::FreeBuffer((void**)&palette);
	bufsize=0;
	width=height=0;
	pixelformat=0;
	nbcolors=0;
	rmask=0xFF;
	gmask=0xFF00;
	bmask=0xFF0000;
	amask=0xFF000000;
}



BOOLEAN TWPICTURE::SetPixelFormat(WORD npf,DWORD nmr,DWORD nmg,DWORD nmb,DWORD nma)
{
	
	if(npf==pixelformat) return TRUE;
//	DWORD col;
	DWORD	nbpix=height*width;
	TWCOLOR c;

	TWPICTURE p;
	p.pixelformat=pixelformat;
	p.buffer=buffer;
	p.width=width;
	p.height=height;
	p.amask=amask;
	p.rmask=rmask;
	p.gmask=gmask;
	p.bmask=bmask;

	
	
	buffer=NULL;
	pixelformat=npf;
	CreateBuffer();
	SetMasks(nmr,nmg,nmb,nma);

	for(DWORD i=0;i<nbpix;i++)
	{
		//on lit la couleur
		c.SetMasks(p.rmask,p.gmask,p.bmask,p.amask);
		p.GetPixel(i,c);
		//et on l'crit
		c.SetMasks(rmask,gmask,bmask,amask);
		SetPixel(i,c);
	}

	p.Release();
	return TRUE;

}


BOOLEAN TWPICTURE::GetPalEntry(DWORD index, TWCOLOR &c)
{
	if((!palette)||(index>=nbcolors)) return FALSE;
	c.Set(palette[(index<<2)+2],palette[(index<<2)+1],palette[(index<<2)],palette[(index<<2)+3]);
	return TRUE;
}

BOOLEAN TWPICTURE::GetPalEntry(DWORD index, DWORD &c)
{
	if((!palette)||(index>=nbcolors)) return FALSE;
	c=palette[(index<<2)+2]|palette[(index<<2)+1]>>8|palette[(index<<2)]>>16;
	return TRUE;
}

void TWPICTURE::CreatePalette(DWORD nbcol)
{
	nbcolors=nbcol;
	TWOBJECT::FreeBuffer((void**)&palette);
	TWOBJECT::CreateBuffer((void**)&palette,nbcol<<2);
}


void TWPICTURE::VFlip()
{
	DWORD linecnt=(width*pixelformat)>>3;
	BYTE * buf=new BYTE[linecnt];
	BYTE *ptr=(BYTE*)buffer,*ptr2=ptr+bufsize-linecnt;
	for(DWORD i=0;i<(height>>1);i++)
	{
		memcpy(buf,ptr,linecnt);
		memcpy(ptr,ptr2,linecnt);
		memcpy(ptr2,buf,linecnt);
		ptr+=linecnt;
		ptr2-=linecnt;
	}
	SAFEDELA(buf);
}

void TWPICTURE::Resize(DWORD nw, DWORD nh)
{
	DWORD *d32,*p32=(DWORD*)buffer;
	WORD *d16,*p16=(WORD*)buffer;
	BYTE *d8,*p8=(BYTE*)buffer;
	DWORD oldw=width,oldh=height;
	DWORD i,j,offss,offsd;
	double k,l,ratiow,ratioh;
	BYTE tmp;
	
	if((nw==(DWORD)width) && (nh==(DWORD)height)) return;
	ratiow=(double)width/(double)nw;
	ratioh=(double)height/(double)nh;

	buffer=NULL;
	width=nw;
	height=nh;
	CreateBuffer();
	d8=(BYTE*)buffer;
	d16=(WORD*)buffer;
	d32=(DWORD*)buffer;

	for(j=0;j<(DWORD)height;j++)
	{
		l=j-(double)((height)>>1);
		l*=ratioh;
		l+=(double)(oldh>>1);

		for(i=0;i<(DWORD)width;i++)
		{
			k=i-(double)((width)>>1);
			k*=ratiow;
			k+=(double)(oldw>>1);

			offss=oldw*(DWORD)l+(DWORD)k;
			offsd=width*j+i;
			switch(pixelformat&0xFF)
			{
			case 1:
				tmp=p8[offss>>3];
				tmp&=1<<(offss%8);
				tmp>>=(offss%8);
				tmp<<=(offsd%8);
				d8[offsd>>3]&=~tmp;
				d8[offsd>>3]|=tmp;

				break;
			case 4:	//a faire
				break;
			case 8:
				d8[offsd]=p8[offss];
				break;
			case 16:
				d16[offsd]=p16[offss];
				break;
			case 24:
				d8[3*offsd]=p8[3*offss];
				d8[3*offsd+1]=p8[3*offss+1];
				d8[3*offsd+2]=p8[3*offss+2];
				break;
			case 32:
				d32[offsd]=p32[offss];
				break;
			}
		}
	}

	SAFEDELA(p8);
}

void TWPICTURE::SetPixel(DWORD i, DWORD c)
{
	TWCOLOR col;
	col.SetRGBA32(c);
	SetPixel(i,col);
}

void TWPICTURE::SetPixel(DWORD i, TWCOLOR &c)
{
	DWORD *p32=(DWORD*)buffer;
	WORD *p16=(WORD*)buffer;
	BYTE *p8=(BYTE*)buffer;

	c.SetMasks(rmask,gmask,bmask,amask);
	switch(pixelformat)
	{
	case PF_MONO:	//a faire
		//if(!Find
		//if(!GetPalEntry(p8[i>>8]&(i%8),&c)) c.Set((DWORD)(p8[i>>8]&(i%8))*0x00FFFFFF);
		break;

	case PF_PAL4:
		//if(!GetPalEntry(p8[i],&c)) c.Set(p8[i],p8[i],p8[i]);
		break;

	case PF_PAL8:
		//if(!GetPalEntry(p8[i],&c)) c.Set(p8[i],p8[i],p8[i]);
		break;

	case PF_RGB16:
		//if(!GetPalEntry(p16[i],&c)) c.Set(p16[i]);
		p16[i]=c.GetRGB16();
	break;

	case PF_RGB24:
		p8[i*3]=c.GetRed();
		p8[i*3+1]=c.GetGreen();
		p8[i*3+2]=c.GetBlue();
		break;

	case PF_RGB32:
		p32[i]=c.GetRGBA32();
		break;
	case PF_YUV24:
		p8[i*3+2]=c.GetY();
		p8[i*3+1]=c.GetU();
		p8[i*3]=c.GetV();
		break;

	case PF_YUV32:
		p32[i]=c.GetYUV32();
		break;
	}
}
DWORD TWPICTURE::GetPixel(DWORD i)
{
	DWORD c,*p32=(DWORD*)buffer;
	WORD *p16=(WORD*)buffer;
	BYTE *p8=(BYTE*)buffer;

	switch(pixelformat)
	{
	case PF_MONO:
		if(!GetPalEntry(p8[i>>8]&(1<<(i%8)),c)) c=((DWORD)(p8[i>>8]&(1<<(i%8)))*0x00FFFFFF);
		break;
	case PF_PAL4:
		if(!this->GetPalEntry(p8[i>>2]&(0xF<<((i%2)<<2)),c)) c=((DWORD)(p8[i>>8]&(0xF<<((i%2)<<2)))*0x00FFFFFF);
		break;
	case PF_PAL8:
		if(!GetPalEntry(p8[i],c)) c=p8[i]|(p8[i]<<8)|(p8[i]<<16);		
		break;
	case PF_RGB16:
		if(!GetPalEntry(p16[i],c)) c=(((p16[i]&0x1F)*8)/6)|((((p16[i]&0x7E0)*8)/6)<<8)|((((p16[i]&0xF800)*8)/5)<<16);
		break;
	case PF_RGB24:
		c=p8[i*3]|p8[i*3+1]<<8|p8[i*3+2]<<16;
		break;
	case PF_YUV24:
		c=p8[i*3]|p8[i*3+1]<<8|p8[i*3+2]<<16;
		break;

	case PF_RGB32:
		c=p32[i];
		break;

	case PF_YUV32:
		c=p32[i];
		break;
	}
	return c;
}

void TWPICTURE::GetPixel(DWORD i, TWCOLOR &c)
{
	DWORD ct=GetPixel(i);
	c.SetRGBA32(ct);
}

void *& TWPICTURE::Create(DWORD wid, DWORD hei, short pf)
{
	Release();
	pixelformat=pf;
	width=wid;
	height=hei;
	return CreateBuffer();
}

void TWPICTURE::Fill(TWCOLOR &c)
{
	for(DWORD i=0;i<(DWORD)(width*height);i++) SetPixel(i,c);
}

void TWPICTURE::BGRToRGB()
{
	TWCOLOR col;
	for(DWORD i=0;i<(width*height);i++)
	{
		GetPixel(i,col);
		col.BGRToRGB();
		SetPixel(i,col);
	}
}

void TWPICTURE::AAliasDouble(int alias)
{
	Resize(width<<1,height<<1);
	switch(alias)
	{
	case 5: AAlias5();
		break;
	case 9: AAlias9();
		break;
	default: AAlias3();
	}
}
void TWPICTURE::AAlias3()
{
	VALUE r,g,b;
	DWORD offs;
	TWCOLOR col;
	for(DWORD i=0;i<height-1;i++)
		for(DWORD j=0;j<width-1;j++)
		{
			offs=width*i+j;
			GetPixel(offs,col);
			col.GetValues(r,g,b);
			GetPixel(offs+1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs+width,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			r/=3;
			g/=3;
			b/=3;
			col.SetValues(r,g,b);
			SetPixel(offs,col);
		}
}

void TWPICTURE::AAlias5()
{
	VALUE r,g,b;
	DWORD offs;
	TWCOLOR col;
	for(DWORD i=1;i<height-1;i++)
		for(DWORD j=1;j<width-1;j++)
		{
			offs=width*i+j;
			GetPixel(offs,col);
			col.GetValues(r,g,b);
			GetPixel(offs+1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs+width,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-width,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			r/=5;
			g/=5;
			b/=5;
			col.SetValues(r,g,b);
			SetPixel(offs,col);
		}
}

void TWPICTURE::AAlias9()
{
	VALUE r,g,b;
	DWORD offs;
	TWCOLOR col;
	for(DWORD i=1;i<height-1;i++)
		for(DWORD j=1;j<width-1;j++)
		{
			offs=width*i+j;
			GetPixel(offs,col);
			col.GetValues(r,g,b);
			GetPixel(offs+1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs+width,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-width,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs+width+1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs+width-1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-width+1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			GetPixel(offs-width-1,col);
			r+=col.GetRedValue();
			g+=col.GetGreenValue();
			b+=col.GetBlueValue();
			r/=9;
			g/=9;
			b/=9;
			col.SetValues(r,g,b);
			SetPixel(offs,col);
		}
}


/*DWORD TWPICTURE::MarkLine(long x,long y,const TWCOLOR &col,const VALUE &threshold,int dx, int dy)
{
	DWORD notsim=0,res=0,offs=width*y+x;
	if(workbuf[offs]) return 0;
	static TWCOLOR c1;
	DWORD res=0;

	if(!workbuf[offs]) workbuf[offs]=1;

	x+=dx;
	y+=dy;
	if((x<0)||(x>=width)||(y<0)||(y>=height)) 
	{
		workbuf[offs]=2;
		return 0;
	}
	offs=width*y+x;
	if(workbuf[offs]==1)
	{
		do
		{
			++res;
			x+=dx;y+=dy;
			offs=width*y+x;
		}while((x>=0)&&(x<width)&&(y>=0)&&(y<height)&&(workbuf[offs]==1));
	}

	do
	{
		++res;
		x+=dx;y+=dy;
		offs=width*y+x;
	}while((x>=0)&&(x<width)&&(y>=0)&&(y<height)&&(workbuf[offs]==0));

	if((x<0)||(x>=width)||(y<0)||(y>=height)) 
	{
		x-=dx;
		y-=dy;
		offs=width*y+x;
		workbuf[offs]=2;
		return res;
	}


	offs=(width*y)+x;

	do
	{
		x+=dx;
		y+=dy;
		offs=(width*y)+x;
	} while(workbuf[offs]!=

}
*/
int TWPICTURE::Mark4Linked(TWCOLLECTION <DWORD> & colres,const int x,const int y,const TWCOLOR &col,const VALUE &threshold, short db)
{
	int tres=0,res=0,offs=width*y+x;
	if(workbuf[offs]) return 0;
	static TWCOLOR c1;
	GetPixel(offs,c1);
	if(col.IsSimilar(c1)<threshold) return -1;
	if(db==5) return 1;

	workbuf[offs]=1;
	int i,tx,ty,tx2,ty2;
	for(i=0;i<4;i++)
	{
		tx=x;ty=y;
		switch((i+db)%4)
		{
		case 0: ++tx; break;
		case 1: ++ty; break;
		case 2: --tx; break;
		case 3: --ty; break;
		}
		if((tx<=0) || (ty<=0) || (tx>=(int)width) || (ty>=(int)height))
		{
			if(workbuf[offs]!=2)
			{
				workbuf[offs]=2;
				colres.Add(offs);
				if(x>=(int)width) db=0;
				else if(y>=(int)height) db=1;
				else if(x<0) db=2;
				else if(y<=0) db=3;
			}
		}
		else
		{
			//on vrifie qu'il ne vaut pas mieux tourner a droite
			tx2=x;ty2=y;
			switch((i+db+3)%4)
			{
				case 0: ++tx2; break;
				case 1: ++ty2; break;
				case 2: --tx2; break;
				case 3: --ty2; break;
			}
			if((tx2<=0) || (ty2<=0) || (tx2>=(int)width) || (ty2>=(int)height) || (Mark4Linked(colres,tx2,ty2,col,threshold,5)<0))
				db=(db+i+3)%4;
			else db=(db+i)%4;

			tres=Mark4Linked(colres,tx,ty,col,threshold,db);
			if(tres<0) 
			{
				if(workbuf[offs]!=2)
				{
					workbuf[offs]=2;
					colres.Add(offs);
				}
			}
			else res+=tres;
		}
	}
	
	return res;
}

DWORD TWPICTURE::GetShape(TWCOLLECTION <DWORD> & res, DWORD x, DWORD y, VALUE threshold)
{
	res.Clear();
	if((x>width)||(y>height)) return 0;
	if(!workbuf) CreateWorkBuffer();

	memset(workbuf,0,bufsize);

	DWORD ret=0;
	TWCOLOR col;
	GetPixel(x,y,col);
	
	Mark4Linked(res,x,y,col,threshold);

	return res.GetSize();
}


void TWCOLOR::operator *=(VALUE v)
{
	r=(BYTE)((VALUE)r*v);
	g=(BYTE)((VALUE)g*v);
	b=(BYTE)((VALUE)b*v);
}


TWCOLOR & TWCOLOR::operator +=(const TWCOLOR &c)
{
	int tmp;
	tmp=r+c.r;
	if(tmp>255) tmp=255;
	r=(BYTE)tmp;

	tmp=g+c.g;
	if(tmp>255) tmp=255;
	g=(BYTE)tmp;

	tmp=b+c.b;
	if(tmp>255) tmp=255;
	b=(BYTE)tmp;

	tmp=a+c.a;
	if(tmp>255) tmp=255;
	a=(BYTE)tmp;

	return *this;
}

TWCOLOR & TWCOLOR::operator *=(const TWCOLOR &c)
{
	int tmp;
	tmp=r*c.r;
	if(tmp>255) tmp=255;
	r=(BYTE)tmp;

	tmp=g*c.g;
	if(tmp>255) tmp=255;
	g=(BYTE)tmp;

	tmp=b*c.b;
	if(tmp>255) tmp=255;
	b=(BYTE)tmp;

	tmp=a*c.a;
	if(tmp>255) tmp=255;
	a=(BYTE)tmp;

	return *this;
}

void TWCOLOR::Light(const TWCOLOR &cl)
{
	int tmp;
	tmp=r;
	tmp*=cl.r;
	tmp/=255;
	r=(BYTE)tmp;

	tmp=g;
	tmp*=cl.g;
	tmp/=255;
	g=(BYTE)tmp;

	tmp=b;
	tmp*=cl.b;
	tmp/=255;
	b=(BYTE)tmp;
}

VALUE TWCOLOR::IsSimilar(const TWCOLOR &c) const
{
	VALUE res,tres;
	res=255-ABS(c.r-r);
	res/=1020;
	tres=255-ABS(c.g-g);
	tres/=1020;
	res+=tres;
	tres=255-ABS(c.b-b);
	tres/=1020;
	res+=tres;
	tres=255-ABS(c.a-a);
	tres/=1020;
	res+=tres;
	return res;
}
