#ifndef _TW_2DCORE_H_
#define _TW_2DCORE_H_

#include <twcore.h>
#include <twmodule.h>

#define PF_MONO  0x1
#define PF_PAL4  0x4
#define PF_PAL8	 0x8
#define PF_RGB16 0x10
#define PF_RGB24 0x18
#define PF_RGB32 0x20
#define PF_YUV24 0x118
#define PF_YUV32 0x120


class TWCOLOR;
class TWPICTURE;

/////////////////////////////////////////////////////////////////////
//------------------------ TWCOLOR --------------------------------//
/////////////////////////////////////////////////////////////////////
class TWCOLOR : public TWOBJECT
{
protected:
	BYTE r,g,b,a;
	DWORD rmask,gmask,bmask,amask;

public:
	VALUE IsSimilar(const TWCOLOR &c) const;
	static BYTE		GetComponent(const DWORD & col,DWORD mask);
	static void		SetComponent(DWORD & col,DWORD mask,BYTE comp);

	TWCOLOR(BYTE re,BYTE gr,BYTE bl,BYTE al):TWOBJECT("TWCOLOR") {Set(re,gr,bl,al);SetMasks(0xFF,0xFF00,0xFF0000,0xFF000000);}
	TWCOLOR():TWOBJECT("TWCOLOR") {Set(0,0,0,0);SetMasks(0xFF,0xFF00,0xFF0000,0xFF000000);}
	

	void			BGRToRGB();
	void			Light(const TWCOLOR &cl);

	void			SetYUV(BYTE y,BYTE u,BYTE v,BYTE al=0);

	void operator *=(VALUE v);
	TWCOLOR & operator +=(const TWCOLOR &c);
	TWCOLOR & operator *=(const TWCOLOR &c);

	

	BYTE GetY() const;
	BYTE GetU() const;
	BYTE GetV() const;

	void SetRGBA32(const DWORD &c);
	DWORD GetRGBA32() const;
	
	void SetRGB24(const BYTE *c);

	WORD GetRGB16() const;
	void SetRGB16(WORD c);

	void SetYUV32(DWORD c);
	DWORD GetYUV32() const;

	void SetYUV24(DWORD c);
	DWORD GetYUV24() const;

	//inlines
	void			Set(BYTE re,BYTE gr,BYTE bl,BYTE al=0) {r=re;g=gr;b=bl;a=al;}
	DWORD			GetAMask()				{return amask;}
	DWORD			GetRMask()				{return rmask;}
	DWORD			GetGMask()				{return gmask;}
	DWORD			GetBMask()				{return bmask;}
	void			SetMasks(DWORD r, DWORD g,DWORD b, DWORD a) {rmask=r;gmask=g;bmask=b;amask=a;}
	void			SetValues(VALUE re,VALUE gr,VALUE bl,VALUE al=0) {r=(BYTE)(255*re);g=(BYTE)(255*gr);b=(BYTE)(255*bl);a=(BYTE)(255*al);}
	BYTE			GetRed() const {return r;}
	BYTE			GetGreen() const {return g;}
	BYTE			GetBlue() const {return b;}
	BYTE			GetAlpha() const {return a;}

	VALUE			GetRedValue() const {return ((VALUE)r)/255;}
	VALUE			GetGreenValue() const {return ((VALUE)g)/255;}
	VALUE			GetBlueValue() const {return ((VALUE)b)/255;}
	VALUE			GetAlphaValue() const {return ((VALUE)a)/255;}
	void			GetValues(VALUE &a, VALUE &b, VALUE &c) const {a=GetRedValue();b=GetGreenValue();c=GetBlueValue();}
	operator DWORD() {return GetRGBA32();}
	TWCOLOR & operator =(const TWCOLOR &c){Set(c.r,c.g,c.b,c.a);return *this;}
	TWCOLOR & operator =(const DWORD &v) {SetRGBA32(v);return *this;}
};

/////////////////////////////////////////////////////////////////////
//------------------------ TWPICTURE --------------------------------//
/////////////////////////////////////////////////////////////////////
class TWPICTURE : public TWOBJECT, public TWLOADABLEOBJECT
{
protected:
	DWORD				width;
	DWORD				height;
	WORD				pixelformat;
	void *				buffer;
	DWORD *				workbuf;
	DWORD				bufsize;
	BYTE *				palette;
	DWORD				nbcolors;
	DWORD				hppm;
	DWORD				vppm;
	DWORD				rmask,gmask, bmask,amask;

public:
	//return the number of points composing the shape, and all the indexes of these points
	DWORD				GetShape(TWCOLLECTION<DWORD> &res, DWORD x, DWORD y,VALUE threshold=0.9);
	//put 1 into the workbuffer where pixel is 4-connected, and 2 if boundary
	int					Mark4Linked(TWCOLLECTION<DWORD> &colres,int x,int y,const TWCOLOR &col,const VALUE &threshold, short v=0);
	void				BGRToRGB();
	void *	&			CreateBuffer();
	void *				CreateWorkBuffer();
	void				AAlias3();
	void				AAlias5();
	void				AAlias9();
	void				AAliasDouble(int alias=3);
	void				Fill(TWCOLOR &c);
	void * &			Create(DWORD wid, DWORD hei, short pf);
	void				SetPixel(DWORD ind, TWCOLOR &c);
	void				GetPixel(DWORD ind, TWCOLOR &c);
	void				SetPixel(DWORD ind, DWORD c);
	DWORD				GetPixel(DWORD ind);
	void				Resize(DWORD nw, DWORD nh);
	void				VFlip();
	void				CreatePalette(DWORD nbcol);
	BOOLEAN				GetPalEntry(DWORD index, TWCOLOR &c);
	BOOLEAN				GetPalEntry(DWORD index, DWORD &c);
	BOOLEAN				SetPixelFormat(WORD npf,DWORD nmr,DWORD nmg,DWORD nmb,DWORD nma);
	void				Release();

	TWPICTURE();
	~TWPICTURE();



	//=============== Inlines =====================================
	DWORD		GetAMask()				{return amask;}
	DWORD		GetRMask()				{return rmask;}
	DWORD		GetGMask()				{return gmask;}
	DWORD		GetBMask()				{return bmask;}
	void		SetMasks(DWORD r, DWORD g,DWORD b, DWORD a) {rmask=r;gmask=g;bmask=b;amask=a;}

	void		SetHPPM(DWORD v)		{hppm=v;}
	void		SetVPPM(DWORD v)		{vppm=v;}
	DWORD		GetHPPM()				{return hppm;}
	DWORD		GetVPPM()				{return vppm;}
	DWORD		GetWidth()				{return width;}
	DWORD		GetHeight()				{return height;}
	void		SetWidth(DWORD v)		{width=v;}
	void		SetHeight(DWORD v)		{height=v;}
	WORD		GetPixelFormat()		{return pixelformat;}
	void *		GetBuffer()				{return buffer;}
	void *		GetWorkBuffer()			{return workbuf;}
	BYTE *		GetPalette()			{return palette;}
	DWORD		GetNbPaletteEntries()	{return nbcolors;}
	DWORD		GetSize()				{return bufsize;}
	void		SetPixel(DWORD a,DWORD b, TWCOLOR &c)	{SetPixel(width*b+a,c);}
	void		GetPixel(DWORD a,DWORD b, TWCOLOR &c)	{GetPixel(width*b+a,c);}
	DWORD		GetBufferSize()			{return bufsize;}

};



#endif